# Generated by Django 5.1.4 on 2025-01-09 16:20

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flex_report', '0018_alter_tablebutton_title'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='TemplateSavedFilter',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', models.SlugField(max_length=100, verbose_name='Slug')),
                ('title', models.CharField(max_length=100, verbose_name='Title')),
                ('filters', models.JSONField(default=dict, verbose_name='Filters')),
                ('creator', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='saved_filters', to=settings.AUTH_USER_MODEL)),
                ('template', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='saved_filters', to='flex_report.template')),
            ],
            options={
                'verbose_name': 'Template Saved Filter',
                'verbose_name_plural': 'Template Saved Filters',
                'unique_together': {('slug', 'template'), ('title', 'template')},
            },
        ),
    ]



# Generated by Django 5.0 on 2024-02-15 12:31

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flex_report", "0011_alter_template_model_user_path"),
    ]

    operations = [
        migrations.CreateModel(
            name="TableButtonColor",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        default="", max_length=50, unique=True, verbose_name="title"
                    ),
                ),
                ("color", models.CharField(max_length=50, verbose_name="Color")),
            ],
        ),
        migrations.AlterField(
            model_name="tablebutton",
            name="color",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="buttons",
                to="flex_report.tablebuttoncolor",
            ),
        ),
    ]



from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import F
from django.template.defaultfilters import truncatechars
from django.utils.translation import gettext_lazy as _
from django_better_admin_arrayfield.models.fields import ArrayField
from sortedm2m.fields import SortedManyToManyField

from flex_report import BaseDynamicField, dynamic_field, report_model

from .app_settings import app_settings
from .constants import FieldTypes
from .managers import ColumnManager
from .utils import get_column_type, get_view_name_url, is_field_valid, transform_nulls

DatetimeField = app_settings.DATETIME_FIELD


class TablePage(models.Model):
    title = models.CharField(max_length=200, verbose_name=_("Title"))
    url_name = models.CharField(max_length=200, verbose_name=_("URL Name"))

    @property
    def url(self):
        url = get_view_name_url(self.url_name)
        return f"/{url}" if url else ""

    def __str__(self):
        return f"{self.title}"


class Column(models.Model):
    class COLUMN_TYPES(models.TextChoices):
        model = "model", _("Model")
        dynamic = "dynamic", _("Dynamic")

    title = models.CharField(verbose_name=_("Title"), max_length=150, db_index=True)
    searchable = models.BooleanField(default=False)
    creator = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="created_columns",
        blank=True,
        null=True,
    )
    model = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        related_name="flex_columns",
    )
    column_type = models.CharField(
        choices=COLUMN_TYPES.choices,
        verbose_name=_("Column Type"),
        default=COLUMN_TYPES.model,
    )

    objects = ColumnManager()

    def __str__(self):
        return f"{self.model}: {self.title}"

    def get_dynamic_obj(self) -> None | BaseDynamicField:
        if self.column_type != self.COLUMN_TYPES.dynamic:
            return

        return dynamic_field.get_by_slug(self.title)

    def clean(self):
        if (
            (column_type := get_column_type(self.model, self.title))
            == FieldTypes.dynamic
            and (dynamic_model := self.get_dynamic_obj().model)
            and self.model.model_class() != dynamic_model
        ):
            raise ValidationError(
                {
                    "title": _(
                        "This dynamic column has been registered for another model, which is %(title)s."
                    )
                    % {"title": dynamic_model}
                }
            )

        if column_type and not is_field_valid(self.model.model_class(), self.title):
            raise ValidationError(
                {
                    "title": _(
                        "The field name is not valid. It should be a field on the model."
                    )
                }
            )
        if self.searchable and not is_field_valid(
            self.model.model_class(), self.title, as_filter=True
        ):
            raise ValidationError(
                {
                    "searchable": _(
                        "This field is a non-db field and is not allowed to be used for searching."
                    )
                }
            )

    class Meta:
        unique_together = [("model", "title")]


class TableButtonColor(models.Model):
    title = models.CharField(
        verbose_name=_("title"),
        max_length=50,
        default="",
        unique=True,
    )
    color = models.CharField(max_length=50, verbose_name=_("Color"))

    def __str__(self):
        return f"{self.title} - {self.color}"


class TableButton(models.Model):
    title = models.CharField(
        verbose_name=_("title"),
        max_length=255,
        default="",
        unique=True,
    )
    icon = models.CharField(
        verbose_name=_("Icon"),
        max_length=30,
        blank=True,
        null=True,
    )
    display_name = models.CharField(
        verbose_name=_("Display Name"),
        max_length=40,
        blank=True,
        null=True,
    )
    event = models.CharField(
        max_length=50,
        verbose_name=_("Event"),
        blank=True,
        null=True,
    )
    exposed_fields = ArrayField(
        models.CharField(max_length=50),
        verbose_name=_("Exposed Fields"),
        default=list,
        blank=True,
    )
    url_name = models.CharField(
        max_length=200,
        verbose_name=_("URL Name"),
        blank=True,
        null=True,
    )
    url_kwargs = models.JSONField(
        verbose_name=_("URL Parameters"),
        default=dict,
        blank=True,
    )
    query_strings = models.JSONField(
        verbose_name=_("Query String"),
        default=dict,
        blank=True,
    )
    color = models.ForeignKey(
        TableButtonColor,
        on_delete=models.SET_NULL,
        null=True,
        related_name="buttons",
    )
    is_table = models.BooleanField(
        verbose_name=_("Is Table"),
        default=False,
    )

    @property
    def url(self):
        url = get_view_name_url(self.url_name)
        return f"/{url}" if url else ""

    def clean(self):
        if not (self.title or self.icon):
            raise ValidationError({"title": "Title or icon is required."})

        if not (bool(self.event) ^ bool(self.url_name)):
            raise ValidationError(
                {"event": "Filling either of Event or URL Name is required."}
            )

    def __str__(self):
        return (
            f"{self.title} - {self.color.title} -> "
            f"{self.url_name or truncatechars(self.event, 15)}"
        )


@report_model.register
class Template(models.Model):
    class Status(models.TextChoices):
        complete = "c", _("Completed")
        pending = "p", _("Pending")

    title = models.CharField(max_length=200, verbose_name=_("Title"))
    filters = models.JSONField(verbose_name=_("Filters"), default=dict, blank=True)
    columns = SortedManyToManyField(
        Column,
        blank=True,
        limit_choices_to={"model": F("model")},
        verbose_name=_("Columns"),
    )
    buttons = SortedManyToManyField(
        TableButton,
        blank=True,
        related_name="templates",
        verbose_name=_("Buttons"),
    )
    has_export = models.BooleanField(default=True, verbose_name=_("Has Export"))
    model_user_path = models.JSONField(
        max_length=200,
        verbose_name=_("Model User Path"),
        blank=True,
        null=True,
        default=dict,
    )
    creator = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        null=True,
        related_name="templates",
    )
    model = models.ForeignKey(
        ContentType,
        verbose_name=_("Model"),
        on_delete=models.CASCADE,
        related_name="flex_templates",
    )
    page = models.ForeignKey(
        TablePage,
        verbose_name=_("Page"),
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    is_page_default = models.BooleanField(
        verbose_name=_("Page Default"),
        default=False,
    )
    created_date = DatetimeField(
        auto_now_add=True,
        verbose_name=_("Created Date"),
    )
    modified_date = DatetimeField(auto_now=True, verbose_name=_("Modified Date"))
    status = models.CharField(
        max_length=1,
        verbose_name=_("Status"),
        choices=Status.choices,
        default=Status.pending,
    )

    @property
    def is_completed(self):
        return self.status == self.Status.complete

    @property
    def columns_count(self):
        return self.columns.count()

    @property
    def filters_count(self):
        return len(list(filter(None, (self.filters or {}).values())))

    @property
    def user_fullname(self):
        return getattr(self.creator, "full_name", _("Not Set"))

    def __str__(self):
        return self.title

    def get_queryset(self):
        return self.model.model_class()._default_manager.filter(
            **transform_nulls(self.filters)
        )

    class Meta:
        verbose_name = _("Template")
        verbose_name_plural = _("Templates")


class TemplateSavedFilter(models.Model):
    title = models.CharField(max_length=100, verbose_name=_("Title"))
    filters = models.JSONField(verbose_name=_("Filters"), default=dict)
    template = models.ForeignKey(
        Template,
        on_delete=models.CASCADE,
        related_name="saved_filters",
    )
    created = DatetimeField(auto_now_add=True, verbose_name=_("Created"))
    modified = DatetimeField(auto_now=True, verbose_name=_("Modified"))
    creator = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        related_name="saved_filters",
        null=True,
        blank=True,
    )

    def get_queryset(self):
        return self.template.get_queryset().filter(**transform_nulls(self.filters))

    def __str__(self):
        return self.title

    class Meta:
        verbose_name = _("Template Saved Filter")
        verbose_name_plural = _("Template Saved Filters")
        unique_together = [("title", "template")]



# Generated by Django 5.1.4 on 2025-01-10 18:08

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flex_report', '0019_templatesavedfilter'),
    ]

    operations = [
        migrations.AddField(
            model_name='templatesavedfilter',
            name='created',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now, verbose_name='Created'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='templatesavedfilter',
            name='modified',
            field=models.DateTimeField(auto_now=True, verbose_name='Modified'),
        ),
    ]



# Generated by Django 5.0 on 2024-10-19 05:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flex_report", "0017_remove_column_groups_remove_column_users_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="tablebutton",
            name="title",
            field=models.CharField(
                default="", max_length=255, unique=True, verbose_name="title"
            ),
        ),
    ]



# Generated by Django 5.1.5 on 2025-01-19 20:14

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('flex_report', '0023_alter_template_created_date_and_more'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='templatesavedfilter',
            unique_together={('title', 'template')},
        ),
        migrations.RemoveField(
            model_name='templatesavedfilter',
            name='slug',
        ),
    ]



# Generated by Django 5.1.4 on 2025-01-12 21:01

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flex_report', '0020_templatesavedfilter_created_and_more'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='templatesavedfilter',
            unique_together={('title', 'template')},
        ),
        migrations.AlterField(
            model_name='templatesavedfilter',
            name='creator',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='saved_filters', to=settings.AUTH_USER_MODEL),
        ),
        migrations.RemoveField(
            model_name='templatesavedfilter',
            name='slug',
        ),
    ]



from .app_settings import app_settings
from .defaults.views import (
    BaseView,
    ColumnCreateView,
    ColumnDeleteView,
    ColumnListView,
    ColumnUpdateView,
    GeneralQuerySetExportView,
    ReportExportView,
    ReportView,
    TemplateCloneView,
    TemplateCreateCompleteView,
    TemplateCreateInitView,
    TemplateDeleteView,
    TemplateListView,
    TemplateToggleDefaultView,
    TemplateUpdateView,
)

BaseView = app_settings.BASE_VIEW or BaseView


column_create_view = ColumnCreateView.as_view()


column_list_view = ColumnListView.as_view()


column_update_view = ColumnUpdateView.as_view()


column_delete_view = ColumnDeleteView.as_view()


template_list_view = TemplateListView.as_view()


template_delete_view = TemplateDeleteView.as_view()


template_create_init_view = TemplateCreateInitView.as_view()


template_clone_view = TemplateCloneView.as_view()


template_toggle_default_view = TemplateToggleDefaultView.as_view()


template_create_complete_view = TemplateCreateCompleteView.as_view()


template_update_view = TemplateUpdateView.as_view()


report_view = ReportView.as_view()


general_qs_export_view = GeneralQuerySetExportView.as_view()


report_export_view = ReportExportView.as_view()



from django_better_admin_arrayfield.admin.mixins import DynamicArrayMixin

from django import forms
from django.contrib import admin
from django.contrib.contenttypes.models import ContentType
from django.utils.html import format_html

from ..models import (
    Column,
    TableButton,
    TableButtonColor,
    TablePage,
    Template,
    TemplateSavedFilter,
)
from ..utils import fields_join, get_table_page_choices, get_table_page_optional_choices


class TableButtonColorAdmin(admin.ModelAdmin):
    list_display = [
        TableButtonColor.title.field.name,
        TableButtonColor.color.field.name,
    ]
    search_fields = [
        TableButtonColor.title.field.name,
        TableButtonColor.color.field.name,
    ]


class TemplateAdmin(admin.ModelAdmin):
    list_display = [
        Template.title.field.name,
        Template.creator.field.name,
        Template.model.field.name,
        Template.created_date.field.name,
        Template.modified_date.field.name,
        Template.status.field.name,
        Template.is_page_default.field.name,
        Template.columns_count.fget.__name__,
    ]
    raw_id_fields = [
        Template.creator.field.name,
        Template.model.field.name,
    ]
    search_fields = [
        Template.title.field.name,
        fields_join(Template.model.field.name, ContentType.model.field.name),
    ]
    list_filter = [
        Template.is_page_default.field.name,
        Template.status.field.name,
    ]


class TablePageAdmin(admin.ModelAdmin):
    @admin.display(description="URL Name")
    def url(self, obj):
        return format_html(f"<a href='{obj.url}'>{obj.url_name}</a>")

    readonly_fields = (url.__name__,)
    list_display = (TablePage.title.field.name, url.__name__)
    search_fields = readonly_fields + list_display

    def get_form(self, request, obj=None, **kwargs):
        kwargs["widgets"] = {TablePage.url_name.field.name: forms.Select(choices=get_table_page_choices())}
        return super().get_form(request, obj, **kwargs)

    def get_readonly_fields(self, _, obj=None):
        return self.readonly_fields if obj else ()


class ColumnAdmin(admin.ModelAdmin):
    list_display = [
        Column.title.field.name,
        Column.model.field.name,
    ]
    search_fields = [
        Column.title.field.name,
        fields_join(Column.model.field.name, ContentType.model.field.name),
    ]
    list_filter = [
        Column.searchable.field.name,
        Column.column_type.field.name,
    ]


class TableButtonAdmin(admin.ModelAdmin, DynamicArrayMixin):
    @admin.display(description="URL Name")
    def url(self, obj):
        return format_html(f"<a href='{obj.url}'>{obj.url_name}</a>")

    list_readonly_fields = [url.__name__, TableButton.title.field.name]
    list_display = [
        TableButton.title.field.name,
        TableButton.display_name.field.name,
        TableButton.icon.field.name,
        url.__name__,
        TableButton.color.field.name,
    ]
    search_fields = [
        TableButton.url_name.field.name,
        TableButton.title.field.name,
        TableButton.display_name.field.name,
        TableButton.color.field.name,
    ]
    list_editable = list(set(list_display) - set(list_readonly_fields))

    def get_form(self, request, obj=None, **kwargs):
        kwargs["widgets"] = {TableButton.url_name.field.name: forms.Select(choices=get_table_page_optional_choices())}
        return super().get_form(request, obj, **kwargs)

    def get_readonly_fields(self, _, obj=None):
        return self.readonly_fields if obj else ()


class TemplateSavedFilterAdmin(admin.ModelAdmin):
    list_display = [
        TemplateSavedFilter.template.field.name,
        TemplateSavedFilter.title.field.name,
        TemplateSavedFilter.creator.field.name,
        TemplateSavedFilter.created.field.name,
        TemplateSavedFilter.modified.field.name,
    ]
    autocomplete_fields = [
        TemplateSavedFilter.template.field.name,
        TemplateSavedFilter.creator.field.name,
    ]
    search_fields = [
        TemplateSavedFilter.title.field.name,
        fields_join(TemplateSavedFilter.template.field.name, Template.title.field.name),
    ]



from functools import reduce
from operator import or_

import django_filters
from django_filters.filters import LOOKUP_SEP
from django_filters import FilterSet as BaseFilterSetBase

from django_filters.conf import DEFAULTS
from django import forms
from django.db.models import Q
from django.utils.translation import gettext_lazy as _

from .app_settings import app_settings
from .utils import (
    generate_filterset_form,
    get_fields_lookups,
    get_model_filters,
    get_quicksearch_fields_lookups,
    get_annotated_fields_lookups,
    get_annotated_fields,
    get_model_manager,
    get_field,
)


class FilterSetMeta:
    def __init__(self, model, fields=[]):
        self.model = model
        self.searchable_fields = fields
        
        
class FilterSetBase(BaseFilterSetBase):
    def __init__(self, *args, **kwargs):
        queryset = kwargs.get("queryset")
        if not queryset:
            queryset = get_model_manager(self.Meta.model).all()
            
        kwargs.update(queryset=queryset)
        super().__init__(*args, **kwargs)

        
        
class CustomModelMultipleChoiceFilter(django_filters.ModelMultipleChoiceFilter):
    def get_filter_predicate(self, v):
        name = self.field_name
        if name:
            name = LOOKUP_SEP.join([name, self.lookup_expr])
        predicate = super().get_filter_predicate(v)
        predicate[name] = [v]
        return predicate


class FilterSet(FilterSetBase):
    form_classes = []
    
    @classmethod
    def get_fields(cls):
        model = cls._meta.model
        fields, exclude = get_model_filters(model)
        cls._meta.fields = fields and get_fields_lookups(model, fields)
        cls._meta.exclude = exclude and get_fields_lookups(model, exclude)
        
        for annotation_name, field_type in get_annotated_fields(model).items():
            annotation_field = get_field(model, annotation_name).output_field
            for lookup in get_annotated_fields_lookups(model).get(annotation_name, ["exact"]):
                field_name = LOOKUP_SEP.join([annotation_name, lookup])
                defaults = {
                    "field_name": annotation_name,
                    "label": " ".join(str(DEFAULTS.get("VERBOSE_LOOKUPS", {}).get(field_name_part, field_name_part)) for field_name_part in field_name.split(LOOKUP_SEP)).replace("_", " ").title(),
                    "lookup_expr": lookup,
                }
                filter_class, opts = cls.filter_for_lookup(annotation_field, field_type)
                defaults.update(opts)
                cls.declared_filters[field_name] = filter_class(**defaults)
                
        return super().get_fields()

    @classmethod
    def filter_for_lookup(cls, f, lookup_type):
        filter_, opts = super().filter_for_lookup(f, lookup_type)

        match filter_:
            case django_filters.ModelMultipleChoiceFilter if "choices" in opts:
                filter_ = django_filters.MultipleChoiceFilter
                opts.update(widget=filter_.field_class.widget())
            case django_filters.ModelChoiceFilter:
                filter_ = django_filters.ModelMultipleChoiceFilter
                opts.update(widget=filter_.field_class.widget())
            case _ if issubclass(filter_, django_filters.BaseInFilter):
                filter_ = CustomModelMultipleChoiceFilter
                opts.update(widget=filter_.field_class.widget())
        return filter_, opts

    def get_form_class(self):
        base_form_class = super().get_form_class()
        return (
            generate_filterset_form(
                self._meta.model,
                form_classes=self.form_classes + [base_form_class],
            )
            if self.form_classes
            else base_form_class
        )


class QuicksearchFilterset(FilterSet):
    search = django_filters.CharFilter(
        method="quick_search",
        label=_("Quick Search"),
        widget=forms.TextInput(
            attrs={
                "hx-trigger": "keyup changed delay:700ms",
                "onkeypress": "return event.keyCode != 13;",
                "class": "form-control",
            }
        ),
    )

    def quick_search(self, queryset, name, value):
        if not value.strip():
            return queryset
        fields = self.Meta.searchable_fields or []
        q_object = reduce(
            or_, [(Q(**{f"{field}__icontains": value})) for field in fields]
        )
        return queryset.filter(q_object) if q_object else queryset.none()

    @classmethod
    def get_fields(cls):
        model = cls._meta.model
        _, exclude = get_model_filters(model)
        cls._meta.fields = ["search"]
        cls._meta.exclude = exclude and get_quicksearch_fields_lookups(model, exclude)
        return super(FilterSet, cls).get_fields()


def generate_quicksearch_filterset_from_model(model, fields=[], form_classes=None):
    return type(
        f"{getattr(model, '__name__', '')}DynamicQuicksearchFilterset",
        (QuicksearchFilterset,),
        {"Meta": FilterSetMeta(model, fields), "form_classes": form_classes},
    )


def generate_filterset_from_model(model, form_classes=None):
    if form_classes is None:
        form_classes = []

    return type(
        f"{getattr(model, '__name__', '')}DynamicFilterSet",
        (app_settings.FILTERSET_CLASS,),
        {"Meta": FilterSetMeta(model), "form_classes": form_classes},
    )



from django.db import models
from django.utils.translation import gettext_lazy as _


class TemplateTypeChoices(models.TextChoices):
    report = "report", _("This template is related to a report")
    page = "page", _("The template is related to one of the pages in the system")



import contextlib
import json
import math
from operator import call
from typing import Iterable
from urllib.parse import urlencode

from django import template
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models import Model, QuerySet
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch
from django.utils.safestring import mark_safe
from flex_report.app_settings import app_settings
from flex_report.utils import (
    get_col_verbose_name,
    get_column_cell,
    get_model_field,
)

register = template.Library()


@register.simple_tag
def call_method(obj, method_name, *args, **kwargs):
    with contextlib.suppress(AttributeError):
        return call(getattr(obj, method_name), *args, **kwargs)

    return mark_safe("&mdash;")


@register.filter(name="enumerate")
def enum(iterable: Iterable):
    return enumerate(iterable)


@register.inclusion_tag("flex_report/pagination.html", takes_context=True)
def show_pagination(context, pagination_context=None, *, link_attributes=None, scroll_tag=None):
    return {
        "request": context["request"],
        "link_attributes": link_attributes,
        "scroll_tag": scroll_tag,
        **(pagination_context or context["pagination"]),
    }


@register.filter(name="range")
def get_range(value):
    return range(value)


@register.filter
def is_editor(user):
    group, _ = Group.objects.get_or_create(name=app_settings.EDITORS_GROUP_NAME)
    return user.groups.filter(pk=group.pk).exists() or user.is_superuser


@register.filter
def get_verbose_name(obj, plural=False):
    assert hasattr(obj, "_meta") or isinstance(obj, QuerySet)
    if isinstance(obj, Model):
        meta = obj._meta

    elif isinstance(obj, QuerySet):
        meta = obj.model._meta

    else:
        meta = obj._meta.model._meta

    return (meta.verbose_name_plural if plural else meta.verbose_name).title()


@register.filter
def dict_get(dictionary, key):
    return dictionary.get(key)


@register.filter
def get_centered_range(current_page, num_pages):
    max_show_count = 5
    max_count_center = int(max_show_count / 2)
    max_show_count_ceil = math.ceil(max_show_count / 2)
    start = 1
    end = num_pages

    if current_page < max_show_count_ceil < end:
        end = max_show_count if num_pages > max_show_count else num_pages
    elif abs(current_page - num_pages) >= max_show_count_ceil:
        start = current_page - max_count_center
        end = current_page + max_count_center
    elif end > max_show_count:
        start = end - max_show_count

    return range(start, end + 1)


@register.simple_tag
def query_transform(request, **query_string):
    updated = request.GET.copy()
    for k, v in query_string.items():
        updated[k] = v
    return updated.urlencode()


@register.simple_tag(takes_context=True)
def dynamic_query_transform(context, context_name, **kwargs):
    # It's bug. in some contexts dictionary in context is unpacked and we don't need context_name,
    # but some contexts receive original dictionary and passed contexts
    current_context = context.get(context_name) or context
    for k in list(kwargs):
        kwargs[current_context[f"{k}_keyword"]] = kwargs.pop(k)
    return query_transform(context["request"], **kwargs)


def is_row_value_valid(f, v):
    return v or isinstance(f, models.BooleanField) or v == 0


@register.filter
def get_column_verbose_name(model, column):
    return get_col_verbose_name(model, column)


@register.filter
def get_columns_verbose_names(cols: Iterable, obj):
    return list(map(lambda c: get_column_verbose_name(obj, c), cols))


@register.filter
def get_row_value(obj, column):
    field = get_model_field(obj, column)
    value = get_column_cell(obj, column, absolute_url=False)

    tag = app_settings.DATA_TAGS.get(
        type(field),
        app_settings.DATA_TAGS["default"],
    )
    return mark_safe(tag(value) if is_row_value_valid(field, value) else "&mdash;")


@register.inclusion_tag("flex_report/view.html", takes_context=True)
def show_page_report(context):
    return context


@register.simple_tag
def get_report_button_fields(record, button):
    return json.dumps(
        {f: get_column_cell(record, f) for f in button.exposed_fields},
        cls=DjangoJSONEncoder,
    )


@register.simple_tag
def get_report_button_url(record, button):
    url = ""
    url_kwargs = {k: getattr(record, v, v) for k, v in button.url_kwargs.items()}

    with contextlib.suppress(NoReverseMatch):
        url = reverse(
            button.url_name,
            kwargs=url_kwargs | {"ct_pk": record and ContentType.objects.get_for_model(record._meta.model).pk},
        )

    with contextlib.suppress(NoReverseMatch):
        url = reverse(
            button.url_name,
            kwargs=url_kwargs,
        )

    with contextlib.suppress(NoReverseMatch):
        url = reverse(button.url_name)

    url += f"?{urlencode({k: getattr(record, v, v) for k, v in button.query_strings.items()})}"

    return url.strip("?") or "#"



# Generated by Django 5.0 on 2024-02-08 12:34

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flex_report", "0010_alter_column_model_alter_template_model_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="template",
            name="model_user_path",
            field=models.JSONField(
                blank=True,
                default=dict,
                max_length=200,
                null=True,
                verbose_name="Model User Path",
            ),
        ),
    ]



from django.contrib import admin

from .app_settings import app_settings
from .models import (
    Column,
    TableButton,
    TableButtonColor,
    TablePage,
    Template,
    TemplateSavedFilter,
)

for model in [Column, TableButton, TableButtonColor, TablePage, Template, TemplateSavedFilter]:
    admin.site.register(model, app_settings.MODEL_ADMINS[model])



import ast
import contextlib
import mimetypes
from logging import getLogger

from .utils import get_model_manager
from django.core.exceptions import PermissionDenied
from django.core.paginator import EmptyPage, Paginator
from django.http import HttpResponseBadRequest, HttpResponseForbidden
from django.shortcuts import HttpResponse
from django.utils.translation import gettext_lazy as _
from django.views.generic import View
from flex_report import BaseExportFormat, export_format

from .app_settings import app_settings
from .filterset import (
    generate_filterset_from_model,
    generate_quicksearch_filterset_from_model,
)
from .forms import SavedFilterSelectForm
from .models import Template
from .utils import (
    FieldTypes,
    generate_filterset_form,
    get_choice_field_choices,
    get_col_verbose_name,
    get_template_columns,
    string_to_q,
)

logger = getLogger(__file__)


class PaginationMixin(View):
    pages = [25, 75, 100, 200]
    default_page = pages[0]
    pagination = None
    page_keyword = "page"
    per_page_ketyword = "per_page"

    def get_page(self):
        page = self.request.GET.get(self.page_keyword, 1)
        per_page = (
            p
            if (p := self.request.GET.get(self.per_page_ketyword, self.default_page)) and p in map(str, self.pages)
            else self.default_page
        )

        with contextlib.suppress(EmptyPage):
            paginator = Paginator(self.get_paginate_qs(), per_page)
            return paginator.page(page)

        return paginator.page(1)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        page = self.get_page()
        context["pagination"] = self.pagination = {
            "pages": self.pages,
            "qs": page,
            "paginator": page.paginator,
            "page_keyword": self.page_keyword,
            "per_page_keyword": self.per_page_ketyword,
        }
        return context

    def get_paginate_qs(self):
        return []


class TemplateObjectMixin(View):
    template_object = None

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        try:
            self.template_object = self.get_template()
        except PermissionDenied:
            return HttpResponseForbidden()

    def dispatch(self, *args, **kwargs):
        from .models import Template

        handler = None
        match self.template_object and self.template_object.status:
            case Template.Status.complete:
                handler = self.template_ready()
            case Template.Status.pending:
                handler = self.template_not_ready()
        return handler or super().dispatch(*args, **kwargs)

    def get_context_data(self, *args, **kwargs):
        return {
            "realtime_quicksearch": app_settings.REALTIME_QUICKSEARCH,
            "has_export": self.template_object.has_export,
        }

    def get_template(self):
        return self.get_object()

    def template_ready(self):
        pass

    def template_not_ready(self):
        pass


class QuerySetExportMixin(View):
    export_qs = []
    export_headers = {}
    export_columns = []
    export_kwargs = {}
    export_filename = None

    def get_export_filename(self):
        return self.export_filename

    def get_export_columns(self):
        return self.export_columns

    def get_export_headers(self):
        return self.export_headers

    def get_export_qs(self):
        return self.export_qs

    def get_export_kwargs(self):
        return self.export_kwargs

    def get_handle_qs(self):
        return {
            "export_qs": self.get_export_qs(),
            "export_headers": self.get_export_headers(),
            "export_columns": self.get_export_columns(),
            "export_kwargs": self.get_export_kwargs(),
        }

    def check_auth(self):
        if not hasattr((exporter := self.get_exporter()), "check_auth"):
            return

        if exporter.check_auth():
            return

        raise HttpResponseForbidden(content="403 Forbidden")

    def dispatch(self, *args, **kwargs):
        if not (format_ := self.request.GET.get("format", "").lower()) or format_ not in export_format.formats.keys():
            return HttpResponseBadRequest()

        self.export_format = format_
        self.check_auth()

        return super().dispatch(*args, **kwargs)

    def get_exporter(self) -> BaseExportFormat:
        try:
            format_ = export_format.formats[self.export_format]
            if any(self.get_handle_qs().values()):
                return type("DynamicExporter", (format_,), self.get_handle_qs())(
                    request=self.request, user=self.request.user
                )
            return format_(request=self.request, user=self.request.user)
        except KeyError as e:
            raise NotImplementedError(f"The wanted format '{self.export_format}' isn't handled.") from e

    def get(self, *args, **kwargs):
        format_ = self.get_exporter()
        filename = str(format_.get_export_filename())

        response = HttpResponse(
            content_type=mimetypes.types_map.get(
                f".{format_.format_ext}",
                "application/octet-stream",
            ),
            headers={"Content-Disposition": f'attachment; filename="{filename}"'},
        )
        response = format_.handle_response(
            response=response,
        )

        return response


class TablePageMixin(PaginationMixin, TemplateObjectMixin):
    page_keyword = "report_page"
    per_page_keyword = "report_per_page"
    page_template_keyword = "report_template"
    saved_filter_keyword = "saved_filter"

    is_page_table = True
    have_template = True

    template_columns = None
    template_searchable_fields = None
    report_qs = None
    filters = None
    quicksearch = None
    used_filters = None
    ignore_search_values = ["unknown"]
    ignore_search_keys = ["report_template"]

    def get_template(self):
        templates = self.get_page_templates()
        default_template = templates.filter(is_page_default=True).first() or templates.first()
        if not (page_template := self.request.GET.get(self.page_template_keyword)):
            return default_template

        if template := templates.filter(pk=page_template).first():
            return template

        return default_template

    def setup_filters(self):
        saved_filters = getattr(self.get_saved_filter(), "filters", {})
        initials = saved_filters | self.get_initials()
        form_classes = self.get_form_classes()

        self.template_filters = generate_filterset_from_model(self.report_model, form_classes)(
            self.template_object.filters or {}
        )
        self.filters = generate_filterset_from_model(self.report_model, form_classes)(initials)
        self.quicksearch = generate_quicksearch_filterset_from_model(
            self.report_model,
            self.template_searchable_fields.values(),
        )(initials)

    def _has_logical_operator(self, filter_path):
        LOGICAL_OPERATORS = ["(", ")", "&", "|", "!="]
        return any(op in filter_path for op in LOGICAL_OPERATORS)

    def apply_user_path(self, report_qs):
        access_handler = getattr(self.report_model, app_settings.MODEL_USER_PATH_FUNC_NAME, lambda request: {})
        accessed_paths = {
            name: {path: access_handler(self.request).get(name)}
            for name, path in (self.template_object.model_user_path or {}).items()
            if access_handler(self.request).get(name)
        }

        matches_paths = next(iter(accessed_paths.keys()), "__all__")
        if not (accessed_paths and matches_paths):
            return report_qs

        filter_path, filter_value = next(iter(accessed_paths[matches_paths].items()))
        filter_func = self._has_logical_operator(filter_path) and string_to_q or report_qs.filter

        return filter_func(**{filter_path: filter_value}).distinct()

    def _format_used_filter(self, col_name, val):
        formats = {**{k: _("Yes") for k in ["true", "True", True]}, **{k: "خیر" for k in ["false", "False", False]}}
        return formats.get(val, dict(get_choice_field_choices(self.report_model, col_name) or []).get(val, str(val)))

    def used_filter_format(self, col_name, val):
        return (
            ", ".join(map(lambda v: self._format_used_filter(col_name, v), val))
            if isinstance(val, list)
            else self._format_used_filter(col_name, val)
        )

    def validate_filters(self, *filters_list):
        return all(f.get_filters() and f.data and f.is_valid() for f in filters_list)

    def get_saved_filter_form(self):
        return SavedFilterSelectForm(
            self.request.GET,
            template=self.template_object,
        )

    def get_saved_filter(self):
        if not self.get_template():
            return {}

        if not (saved_filter_form := self.get_saved_filter_form()).is_valid():
            return saved_filter_form.initial.get(self.saved_filter_keyword, {})

        return getattr(saved_filter_form, "cleaned_data", {}).get(self.saved_filter_keyword) or {}

    def get_report_qs(self):
        self.setup_filters()

        report_qs = (
            self.template_filters.qs.distinct()
            if self.template_filters.get_filters()
            else get_model_manager(self.report_model).all()
        )
        report_qs = self.apply_user_path(report_qs)

        if self.validate_filters(self.filters, self.quicksearch, self.template_filters):
            report_qs = (
                report_qs.distinct()
                .filter(pk__in=(self.quicksearch.qs.distinct() & self.filters.qs.distinct()).values("pk"))
                .order_by(*self.report_model._meta.ordering or ["pk"])
            )
            cleaned_data = self.quicksearch.form.cleaned_data | self.filters.form.cleaned_data
            self.used_filters = self.get_used_filters(
                {
                    get_col_verbose_name(self.report_model, k): self.used_filter_format(k, v)
                    for k, v in cleaned_data.items()
                    if v
                }
            )

        return report_qs

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)

        if not (obj := self.template_object):
            self.have_template = False
            return

        self.report_model = obj.model.model_class()
        self.template_columns = get_template_columns(obj, as_dict=False)
        self.template_searchable_fields = get_template_columns(obj, searchables=True)
        self.report_qs = self.get_report_qs()

    def get_used_filters(self, cleaned_data):
        return _(" and ").join(
            f'{k} = {", ".join(map(str, v)) if isinstance(v, list) else v}'
            for k, v in cleaned_data.items()
            if k.lower() != "search"
        )

    def _prepare_initial(self, initial):
        if initial.lower() in ["true", "false"]:
            return initial.lower() == "true"
        if initial.startswith("[") and initial.endswith("]") or (initial.isnumeric() and not initial.startswith("0")):
            return ast.literal_eval(initial)
        return initial

    def get_initial_value(self, initial, *, key=""):
        return (
            list(map(self._prepare_initial, self.request.GET.getlist(key)))
            if key.endswith("__in")
            else self._prepare_initial(str(initial))
        )

    def get_initials(self):
        return {
            k: self.get_initial_value(v, key=k)
            for k, v in self.request.GET.dict().items()
            if v.strip() and v not in self.ignore_search_values and k not in self.ignore_search_keys
        }

    def get_form_classes(self):
        if not self.template_object:
            return []
        return [generate_filterset_form(self.report_model)]

    def get_paginate_qs(self):
        return self.get_report_qs()

    def get_context_data(self, **kwargs):
        if not self.have_template:
            return super(TemplateObjectMixin, self).get_context_data(**kwargs)

        context = super().get_context_data(**kwargs)
        context["report"] = {
            "columns": self.template_columns,
            "columns_count": len(self.template_columns)
            + self.template_object.buttons.count()
            + sum(
                len(field.get_dynamic_obj().unpack_field())
                for field in self.template_columns.filter(column_type=FieldTypes.dynamic).only("pk")
            )
            + 1,
            "filters": self.filters,
            "buttons": self.template_object.buttons.all(),
            "searchable_fields": self.template_searchable_fields,
            "quicksearch": self.quicksearch,
            "used_filters": self.used_filters,
            "template": self.template_object,
            "templates": self.get_page_templates(),
            "initials": self.get_initials(),
            "pagination": self.pagination,
            "page_template_keyword": self.page_template_keyword,
            "saved_filter_form": self.get_saved_filter_form(),
            "is_page_table": self.is_page_table,
            "have_template": self.have_template,
            "export_formats": [
                {"name": format_.format_name, "slug": format_.format_slug} for format_ in export_format.formats.values()
            ],
            "page_title": getattr(self.template_object.page, "title", self.template_object.title),
        }
        return context

    def get_page_templates(self):
        return Template.objects.filter(page__url_name=self.request.resolver_match.view_name).order_by(
            "-is_page_default"
        )



# Generated by Django 5.1.4 on 2025-01-13 12:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flex_report', '0022_alter_templatesavedfilter_unique_together_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='template',
            name='created_date',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Created Date'),
        ),
        migrations.AlterField(
            model_name='template',
            name='modified_date',
            field=models.DateTimeField(auto_now=True, verbose_name='Modified Date'),
        ),
    ]



# Generated by Django 5.1.4 on 2025-01-12 21:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flex_report', '0021_alter_templatesavedfilter_unique_together_and_more'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='templatesavedfilter',
            unique_together={('title', 'template')},
        ),
        migrations.AddField(
            model_name='templatesavedfilter',
            name='slug',
            field=models.SlugField(default='default', max_length=100, verbose_name='Slug'),
            preserve_default=False,
        ),
        migrations.AlterUniqueTogether(
            name='templatesavedfilter',
            unique_together={('slug', 'template'), ('title', 'template')},
        ),
    ]
